import numpy as np
import pandas as pd

import tensorflow as tf
import networkx as nx

import math

from scipy.sparse import spdiags
from scipy.sparse.linalg import spsolve
from random import random
from tensorflow.keras.layers import Embedding
from tensorflow.keras import regularizers

import nltk
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords

from collections import Counter
import datetime
import time

nltk.download('punkt_tab')
nltk.download('stopwords')

# reading Reseed data
print('reading Reseed data')
data_clean = pd.read_csv('flickr_data.csv')

all_words = []
for i,row in data_clean.iterrows():
    all_words.extend(word_tokenize(row['title']))
    all_words.extend(word_tokenize(row['tag']))

count = Counter(all_words)

word_list = []
stop_words = set(stopwords.words('english'))
for word in count:
    if not word.isalpha():
        continue
    if len(word) < 2:
        continue
    if word in stop_words:
        continue
    if count[word] < 100:
        continue
    word_list.append(word)

print('number of selected words:', len(word_list))

day_to_dt = {}
start_date = datetime.date(2007, 1, 1)
for i,row in data_clean.iterrows():
    if (i+1) % 10000 == 0:
        print('processing:', (i+1), '/', len(data_clean))
    date_data = pd.to_datetime(row['date'])
    day = (date_data.to_pydatetime().date() - start_date).days // 14
    if day<0 or day > 120:
        continue
    text1 = [w for w in word_tokenize(row['title']) if w in word_list]
    text2 = [w for w in word_tokenize(row['tag']) if w in word_list]
    text = text1 + text2
    if len(text) < 2:
        continue
    if not day in day_to_dt:
        day_to_dt[day] = []
    day_to_dt[day].append(text)
print('data has', len(day_to_dt), 'time units')

# reading ground truth data
print('reading ground truth data')
ground_data = pd.read_csv('reseed_events_upcoming.csv', error_bad_lines=False)

day_to_event_words = {}
date_begin_count = datetime.date(2007, 1, 1)
for i,row in ground_data.iterrows():
    if np.isnan(row['startdate']) or np.isnan(row['enddate']):
        continue
    start_date = datetime.datetime.fromtimestamp(row['startdate']).date()
    end_date = datetime.datetime.fromtimestamp(row['enddate']).date()
    start_day = (start_date - date_begin_count).days
    if start_day > 121*14:
        continue
    end_day = (end_date - date_begin_count).days
    if end_day < 0:
        continue
    if not isinstance(row['description'], str):
        continue
    tks = word_tokenize(row['description'])
    words = [w for w in tks if w in word_list]
    if len(words) < 2:
        continue
    
    days = range(max(start_day//14,0), min(end_day//14,120)+1)
    for day in days:
        if not day in day_to_event_words:
            day_to_event_words[day] = []
        day_to_event_words[day].extend(words)
        
# run GTMA
def dt_to_adj(dt):
    adj = np.zeros((len(word_list), len(word_list)))

    for words in dt:
        for i in range(len(words) - 1):
            for j in range(i+1, len(words)):
                adj[word_list.index(words[i]),word_list.index(words[j])] += 1
                adj[word_list.index(words[j]),word_list.index(words[i])] += 1
    
    return adj

def dt_to_x(dt):
    tf = np.zeros((len(dt), len(word_list)))
    df = np.zeros(len(word_list))
    for i,words in enumerate(dt):
        for word in words:
            tf[i,word_list.index(word)] += 1
        for word in set(words):
            df[word_list.index(word)] += 1
    idf = np.zeros(len(word_list))
    for j in range(len(word_list)):
        if df[j] == 0:
            continue
        idf[j] = math.log(len(dt)/df[j])
    x = np.zeros((len(dt), len(word_list)))
    for i in range(len(dt)):
        for j in range(len(word_list)):
            if tf[i,j] == 0:
                continue
            x[i,j] = (1+math.log2(tf[i,j])) * math.log2(idf[j])
    return x

def softmax_large(v, scale):
    return tf.nn.softmax((v - tf.reduce_max(v, axis=0))/scale, axis=0)

class GTMA(tf.keras.Model):
    def __init__(self, num_nodes, latent_dim, N):
        super(GTMA, self).__init__()
        self.num_nodes = num_nodes
        self.latent_dim = latent_dim
        self.N = N

        # Encoder layers
        self.encoder = tf.keras.Sequential([
            tf.keras.layers.Dense(64, activation='relu'),
            tf.keras.layers.Dense(self.latent_dim, activation='relu')
        ])

        # Decoder layers
        self.decoder = tf.keras.Sequential([
            tf.keras.layers.Dense(64, activation='relu'),
            tf.keras.layers.Dense(self.num_nodes, activation='sigmoid')
        ])

    def build(self, input_shape):
        self.W = self.add_weight(
            shape=(D, self.latent_dim),
            initializer=tf.keras.initializers.RandomUniform(),
            trainable=True,
            name="W"
        )
        self.H = self.add_weight(
            shape=(self.latent_dim, D),
            initializer=tf.keras.initializers.RandomUniform(),
            trainable=True,
            name="H"
        )

    def call(self, data, adjacency_matrix):
        # Encode the input graph
        latent_vector = self.encoder(adjacency_matrix)
        
        # Decode the latent vector to reconstruct the graph
        reconstructed_matrix = self.decoder(latent_vector)

        nmf_result = tf.matmul(tf.matmul(data, self.W), self.H)
        
#         H_out = softmax_large(latent_vector + tf.transpose(self.H), 100)
        
        H_gae = softmax_large(latent_vector, 100)
        H_nmf = softmax_large(tf.transpose(self.H), 100)
        H_out = H_gae + H_nmf
        return reconstructed_matrix, nmf_result, H_out
def run_gtma(day, H_old):
    begin = time.time()

    adj = dt_to_adj(day_to_dt[day])
    D = adj.shape[0]
    
    data = dt_to_x(day_to_dt[day])
    N = data.shape[0]
    
    A = adj

    gtma = GTMA(D, K, N)

    # Defining loss and optimizer
    optimizer = tf.keras.optimizers.Adam(learning_rate=1e-3)
    loss_fn = tf.keras.losses.MeanSquaredError()
    gamma = 0
    alpha = 1
    
    def nmf_loss(X, W, H):
        return tf.reduce_sum(tf.square(X - tf.matmul(W, H)))
    
    # Training loop
    H = None
    for epoch in range(500):
        with tf.GradientTape() as tape:
            gae_A, nmf_A, H = gtma(data, A)
            loss_gae = loss_fn(A, gae_A)
            loss_nmf = loss_fn(data, nmf_A)
            
            smooth_loss = 0
            if day >1:
                H_new = H.numpy().T
                smooth_loss = gamma * np.linalg.norm(H_new - H_old, ord=2) / (D*D)
            
            loss = loss_gae + loss_nmf + smooth_loss
                
        gradients = tape.gradient(loss, gtma.trainable_variables)
        optimizer.apply_gradients(zip(gradients, gtma.trainable_variables))

        if epoch % 100 == 0:
            print('Epoch: {}. Loss: {:.4f}, GAE Loss:{:.6f}, NMF Loss:{:.6f}, smoothing loss:{:.6f}'.format(epoch, loss, loss_gae, loss_nmf, smooth_loss))

    print("Training finished! time: ", time.time()-begin)
    
    H = H.numpy().T

    return H

def print_count(count):
    wc = count.most_common()
    for (w, c) in wc[:10]:
        print('{}\t{}'.format(w, c))

def eval(L, day_to_H, example=False):
    
    prec = 0
    recall = 0
    TP = 0
    PredN = 0
    GroundN = 0
    TP_w = []
    FP_w = []
    FN_w = []
    for day in day_to_event_words:
        if day == 0:
            continue
        ground = day_to_event_words[day]
        H = day_to_H[day]
        for i in range(K):
            topic_score = H[i,:]
            sorted_idx = sorted(range(len(topic_score)), key=lambda k: topic_score[k], reverse=True)
            pred = [word_list[j] for j in sorted_idx[:L]]
            TP_w.extend(list(set(pred).intersection(set(ground))))
            FP_w.extend(list(set(pred) - set(ground)))
            FN_w.extend(list(set(ground) - set(pred)))
            
            TP += len(set(pred).intersection(set(ground)))
            PredN += len(set(pred))
        GroundN += len(set(ground))
    if example:
        print('TP')
        print_count(Counter(TP_w))
        print('FP')
        print_count(Counter(FP_w))
        print('FN')
        print_count(Counter(FN_w))
    
    prec = TP/PredN
    recall = TP/GroundN
    return prec, recall
    
K=10
D=len(word_list)

min_day = 1
max_day = 121
delta = 0.01
day_to_H = {}
for day in range(min_day, max_day):
    print('==== day {} ===='.format(day))
    if day > 1:
        H = run_gtma(day, day_to_H[day-1])
    else:
        H = run_gtma(day, None)

    for i in range(K):
        topic_score = H[i,:]
        sorted_idx = sorted(range(len(topic_score)), key=lambda k: topic_score[k], reverse=True)
        print([word_list[j] for j in sorted_idx[:20]])
    day_to_H[day] = H

Ls = [5, 10, 20, 50, 100]
precs = []
recalls = []
for L in Ls:
    prec, recall = eval(L, day_to_H)
    precs.append(prec)
    recalls.append(recall)
res = precs + recalls

print('finished running')
print('accuracy results')
print('precision at', Ls, ' and recall at', Ls)
print('\t'.join([str(v) for v in precs])+'\t'+'\t'.join([str(v) for v in recalls]))
